/*----------------------------------------------------------------------
    Image filter functions                    filters
        Written by  H.Goto, Mar. 1999
        Modified by H.Goto, June 1999
        Modified by H.Goto, Sep. 1999
----------------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1999  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


#define		GrayLevels	256

#include        <stdio.h>
#include        <stdlib.h>
#include        <string.h>

#include	"filters.h"


static int	mask[] = { 1,1,1, 1,4,1, 1,1,1 };
static int	maskdiv = 12;




/*------------------------------------------------------
	Image Smoothing (color)
------------------------------------------------------*/

int smooth_image_c(SIPImage *src,SIPImage *dst){
	int	x,y,c_off;
	uchar	*pu,*p,*pd;
	uchar	*dp;
	int	sum;
	SIPImage	*tmpimage;
	int	width,height;
	width = src->width;
	height = src->height;
	if ( 0 == (tmpimage = sip_CreateImage(width +2, height +2, 8)) ){
		return(-1);
	}

	for ( c_off = 0 ; c_off < 3 ; c_off++ ){	/* R,G,B order */
		/* copy source image to tmpimage */
		for ( y=0 ; y<height ; y++ ){
			dp = (uchar *)sip_getimgptr(tmpimage,y+1);
			p  = (uchar *)sip_getimgptr(src,y);
			for ( x=0 ; x < width ; x++ )  dp[x+1] = p[4*x + c_off];
			dp[0] = dp[1];  dp[width+1] = dp[width];
		}
		memcpy((void *)sip_getimgptr(tmpimage,0),\
			(void *)sip_getimgptr(tmpimage,1),width +2);
		memcpy((void *)sip_getimgptr(tmpimage,height +1),\
			(void *)sip_getimgptr(tmpimage,height),width +2);

		for ( y=0 ; y<height ; y++ ){
			pu = (uchar *)sip_getimgptr(tmpimage,y);
			p  = (uchar *)sip_getimgptr(tmpimage,y+1);
			pd = (uchar *)sip_getimgptr(tmpimage,y+2);
			dp = (uchar *)sip_getimgptr(dst,y);
			for ( x=0 ; x < dst->width ; x++ ){
				sum  = (int)((uint)pu[x  ]) * mask[0];
				sum += (int)((uint)pu[x+1]) * mask[1];
				sum += (int)((uint)pu[x+2]) * mask[2];
				sum += (int)((uint)p [x  ]) * mask[3];
				sum += (int)((uint)p [x+1]) * mask[4];
				sum += (int)((uint)p [x+2]) * mask[5];
				sum += (int)((uint)pd[x  ]) * mask[6];
				sum += (int)((uint)pd[x+1]) * mask[7];
				sum += (int)((uint)pd[x+2]) * mask[8];
				sum /= maskdiv;
				if ( sum < 0   )  sum = 0;
				if ( sum >= GrayLevels )  sum = GrayLevels -1;
				dp[4*x + c_off] = (uchar)sum;
			}
		}
	}

	sip_DestroyImage(tmpimage);
	return(0);
}




/*------------------------------------------------------
	Image Smoothing
------------------------------------------------------*/

int smooth_image(SIPImage *src,SIPImage *dst){
	int	x,y;
	uchar	*pu,*p,*pd;
	uchar	*dp;
	int	sum;
	SIPImage	*tmpimage;
	int	width,height;

	if ( src->depth != 8 ){
		return( smooth_image_c(src,dst) );
	}

	width = src->width;
	height = src->height;
	if ( 0 == (tmpimage = sip_CreateImage(width +2, height +2, 8)) ){
		return(-1);
	}

	/* copy source image to tmpimage */
	for ( y=0 ; y<height ; y++ ){
		p = (uchar *)sip_getimgptr(tmpimage,y+1);
		memcpy((void *)(p+1),(void *)sip_getimgptr(src,y),width);
		p[0] = p[1];  p[width+1] = p[width];
		if ( y == 0 ){
			memcpy((void *)sip_getimgptr(tmpimage,0),\
				(void *)p,width +2);
		}
	}
	memcpy((void *)sip_getimgptr(tmpimage,height +1),(void *)p,width +2);

	for ( y=0 ; y<height ; y++ ){
		pu = (uchar *)sip_getimgptr(tmpimage,y);
		p  = (uchar *)sip_getimgptr(tmpimage,y+1);
		pd = (uchar *)sip_getimgptr(tmpimage,y+2);
		dp = (uchar *)sip_getimgptr(dst,y);
		for ( x=0 ; x < dst->width ; x++ ){
			sum  = (int)((uint)pu[x  ]) * mask[0];
			sum += (int)((uint)pu[x+1]) * mask[1];
			sum += (int)((uint)pu[x+2]) * mask[2];
			sum += (int)((uint)p [x  ]) * mask[3];
			sum += (int)((uint)p [x+1]) * mask[4];
			sum += (int)((uint)p [x+2]) * mask[5];
			sum += (int)((uint)pd[x  ]) * mask[6];
			sum += (int)((uint)pd[x+1]) * mask[7];
			sum += (int)((uint)pd[x+2]) * mask[8];
			sum /= maskdiv;
			if ( sum < 0   )  sum = 0;
			if ( sum >= GrayLevels )  sum = GrayLevels -1;
			dp[x] = (uchar)sum;
		}
	}
	sip_DestroyImage(tmpimage);
	return(0);
}




/*------------------------------------------------------
	Edge detection by Sobel filter (color)
------------------------------------------------------*/

static int edge_sobel_c(SIPImage *src,SIPImage *dst,int div){
		/* source image      : 32bit color */
		/* destination image : 8bit gray scale */
	int	x,y,x1;
	uchar	*pu,*p,*pd;
	uchar	*dp;
	ulong	*lp;
	int	xsum,ysum;
	ulong	EPW;
	SIPImage	*tmpimage;
	int	width,height;
	width = src->width;
	height = src->height;
	if ( 0 == (tmpimage = sip_CreateImage(width +2, height +2, 32)) ){
		return(-1);
	}

	/* copy source image to tmpimage */
	for ( y=0 ; y<height ; y++ ){
		lp = (ulong *)sip_getimgptr(tmpimage,y+1);
		memcpy((void *)(lp+1),(void *)sip_getimgptr(src,y),width * 4);
		lp[0] = lp[1];  lp[width+1] = lp[width];
		if ( y == 0 ){
			memcpy((void *)sip_getimgptr(tmpimage,0),\
				(void *)lp,(width +2) * 4);
		}
	}
	memcpy((void *)sip_getimgptr(tmpimage,height +1),(void *)lp,(width +2) * 4);

	for ( y=0 ; y<height ; y++ ){
		pu = (uchar *)sip_getimgptr(tmpimage,y);
		p  = (uchar *)sip_getimgptr(tmpimage,y+1);
		pd = (uchar *)sip_getimgptr(tmpimage,y+2);
		dp = (uchar *)sip_getimgptr(dst,y);
		for ( x1=0 ; x1 < dst->width ; x1++ ){
			/* Red */
			x = 4 * x1;
			xsum  = -(int)((uint)pu[x]) +(int)((uint)pu[x+8]);
			xsum += -2 * (int)((uint)p[x]) \
			     + 2 * (int)((uint)p[x+8]);
			xsum += -(int)((uint)pd[x]) + (int)((uint)pd[x+8]);
			ysum  = -(int)((uint)pu[x]) \
			     - 2 * (int)((uint)pu[x+4]) \
			     - (int)((uint)pu[x+8]);
			ysum += (int)((uint)pd[x]) \
			     + 2 * (int)((uint)pd[x+4]) \
			     + (int)((uint)pd[x+8]);
			xsum = xsum * xsum + ysum * ysum;
//			EPW = (ulong)sqrt((double)xsum);
			EPW = (ulong)xsum;

			/* Green */
			x++;
			xsum  = -(int)((uint)pu[x]) +(int)((uint)pu[x+8]);
			xsum += -2 * (int)((uint)p[x]) \
			     + 2 * (int)((uint)p[x+8]);
			xsum += -(int)((uint)pd[x]) + (int)((uint)pd[x+8]);
			ysum  = -(int)((uint)pu[x]) \
			     - 2 * (int)((uint)pu[x+4]) \
			     - (int)((uint)pu[x+8]);
			ysum += (int)((uint)pd[x]) \
			     + 2 * (int)((uint)pd[x+4]) \
			     + (int)((uint)pd[x+8]);
			xsum = xsum * xsum + ysum * ysum;
//			EPW += (ulong)sqrt((double)xsum);
			EPW += (ulong)xsum;

			/* Blue */
			x++;
			xsum  = -(int)((uint)pu[x]) +(int)((uint)pu[x+8]);
			xsum += -2 * (int)((uint)p[x]) \
			     + 2 * (int)((uint)p[x+8]);
			xsum += -(int)((uint)pd[x]) + (int)((uint)pd[x+8]);
			ysum  = -(int)((uint)pu[x]) \
			     - 2 * (int)((uint)pu[x+4]) \
			     - (int)((uint)pu[x+8]);
			ysum += (int)((uint)pd[x]) \
			     + 2 * (int)((uint)pd[x+4]) \
			     + (int)((uint)pd[x+8]);
			xsum = xsum * xsum + ysum * ysum;
//			EPW += (ulong)sqrt((double)xsum);
			EPW += (ulong)xsum;
			EPW = (ulong)sqrt((double)EPW);

			EPW /= div;
			if ( EPW >= GrayLevels )  EPW = GrayLevels -1;
			dp[x1] = (uchar)EPW;
		}
	}
	sip_DestroyImage(tmpimage);
	return(0);
}




/*------------------------------------------------------
	Edge detection by Sobel filter
------------------------------------------------------*/

int edge_sobel(SIPImage *src,SIPImage *dst,int div){
	int	x,y;
	uchar	*pu,*p,*pd;
	uchar	*dp;
	int	xsum,ysum;
	SIPImage	*tmpimage;
	int	width,height;

	if ( src->depth != 8 ){
		return( edge_sobel_c(src,dst,div) );
	}

	width = src->width;
	height = src->height;
	if ( 0 == (tmpimage = sip_CreateImage(width +2, height +2, 8)) ){
		return(-1);
	}

	/* copy source image to tmpimage */
	for ( y=0 ; y<height ; y++ ){
		p = (uchar *)sip_getimgptr(tmpimage,y+1);
		memcpy((void *)(p+1),(void *)sip_getimgptr(src,y),width);
		p[0] = p[1];  p[width+1] = p[width];
		if ( y == 0 ){
			memcpy((void *)sip_getimgptr(tmpimage,0),\
				(void *)p,width +2);
		}
	}
	memcpy((void *)sip_getimgptr(tmpimage,height +1),(void *)p,width +2);

	for ( y=0 ; y<height ; y++ ){
		pu = (uchar *)sip_getimgptr(tmpimage,y);
		p  = (uchar *)sip_getimgptr(tmpimage,y+1);
		pd = (uchar *)sip_getimgptr(tmpimage,y+2);
		dp = (uchar *)sip_getimgptr(dst,y);
		for ( x=0 ; x < dst->width ; x++ ){
			xsum  = -(int)((uint)pu[x]) +(int)((uint)pu[x+2]);
			xsum += -2 * (int)((uint)p[x]) \
			     + 2 * (int)((uint)p[x+2]);
			xsum += -(int)((uint)pd[x]) + (int)((uint)pd[x+2]);
			ysum  = -(int)((uint)pu[x]) \
			     - 2 * (int)((uint)pu[x+1]) \
			     - (int)((uint)pu[x+2]);
			ysum += (int)((uint)pd[x]) \
			     + 2 * (int)((uint)pd[x+1]) \
			     + (int)((uint)pd[x+2]);
			xsum = xsum * xsum + ysum * ysum;
			xsum = (int)sqrt((double)xsum);
			xsum /= div;
			if ( xsum >= GrayLevels )  xsum = GrayLevels -1;
			dp[x] = (uchar)xsum;
		}
	}
	sip_DestroyImage(tmpimage);
	return(0);
}






