/*----------------------------------------------------------------------
    Class for Index Set          iset.cxx
        Written  by H.Goto, Dec. 1997
        Modified by H.Goto, Apr. 1999
----------------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1997-1999  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


#define		IDXSET_ALLOC_STEP	16

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#include	"iset.h"

/*--------------------------------------------------------------
    Index Set
--------------------------------------------------------------*/

IDXSET :: IDXSET(){
	bufsize = counts = actcounts = 0;
}


IDXSET :: ~IDXSET(){
	clear();
}


int IDXSET :: clear(){
	if ( bufsize )  delete []idlist;
	bufsize = counts = actcounts = 0;
}


int IDXSET :: realloc(int size){
	int	i,*tmplist;
	if ( size <= bufsize )  return(0);
	if ( 0 == (tmplist = new int[size]) )  return(-1);
	for ( i=0 ; i<counts ; i++ )  tmplist[i] = idlist[i];
	for ( i=counts ; i<size ; i++ )  tmplist[i] = -1;	/* -1: not used */
	if ( bufsize )  delete []idlist;
	idlist = tmplist;
	bufsize = size;
	return(0);
}


int IDXSET :: addmember(int id){
	if ( bufsize <= counts +1 ){
		if ( realloc(bufsize + IDXSET_ALLOC_STEP) )  return(-1);
	}
	idlist[counts++] = id;
	actcounts++;
	return(counts -1);
}


int IDXSET :: removemember(int i){
	if ( i >= counts )  return(-1);
	if ( idlist[i] != -1 ){  idlist[i] = -1;  actcounts--; }
	return(actcounts);
}


int IDXSET :: truncate(){
	int	i,ncounts;
	for ( i=ncounts=0 ; i<counts ; i++ ){
		if ( idlist[i] != -1 )  idlist[ncounts++] = idlist[i];
	}
	return( counts = ncounts );
}


int IDXSET :: merge(IDXSET *idxset){
	int	i;
	for ( i=0 ; i < idxset->counts ; i++ ){
		if ( idxset->idlist[i] == -1 )  continue;
		if ( 0 > addmember(idxset->idlist[i]) )  return(-1);
	}
	return(0);
}


