/*----------------------------------------------------------------------
    Label Set Control        labelset.cxx
	Written  by H.Goto, Jan. 1998
	Modified by H.Goto, Apr. 1999
----------------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1998,1999  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/

#include	<stdio.h>
#include	<stdlib.h>

#include	"iset.h"
#include	"labelset.h"


/*--------------------------------------------------------------
    Label Set Class
--------------------------------------------------------------*/

LSETS :: LSETS(){
	tblsize = tblsize_rev = counts = 0;
}


LSETS :: ~LSETS(){
	clear();
}


int LSETS :: clear(){
	int	i;
	if ( tblsize ){
		for ( i=0 ; i<tblsize ; i++ ){
			if ( settbl[i] )  delete settbl[i];
		}
		delete []settbl;
	}
	if ( tblsize_rev ){
		delete []revtbl;
	}
	tblsize = tblsize_rev = counts = 0;
	return(0);
}


int LSETS :: settbl_realloc(int size){
	int	i;
	IDXSET	**tmpbuf;
	if ( size <= tblsize )  return(0);
	if ( 0 == (tmpbuf = new IDXSET*[size]) )  return(-1);
	if ( tblsize ){
		memcpy((void *)tmpbuf,(void *)settbl,sizeof(IDXSET*) * counts);
		delete []settbl;
	}
	for ( i=counts ; i<size ; i++ )  tmpbuf[i] = 0;
	settbl = tmpbuf;
	tblsize = size;
	return(0);
}


int LSETS :: revtbl_realloc(int size){
	int	i;
	int	*tmpbuf;
	if ( size <= tblsize_rev )  return(0);
	if ( 0 == (tmpbuf = new int[size]) )  return(-1);
	if ( tblsize_rev ){
		memcpy((void *)tmpbuf,(void *)revtbl,sizeof(int) * tblsize_rev);
		delete []revtbl;
	}
	for ( i=tblsize_rev ; i<size ; i++ )  tmpbuf[i] = -1;	/* not linked */
	revtbl = tmpbuf;
	tblsize_rev = size;
	return(0);
}


int LSETS :: newset(){
	if ( counts + 1 >= tblsize ){
		if ( settbl_realloc(tblsize + 256) )  return(-1);
	}
	if ( 0 == (settbl[counts] = new IDXSET) )  return(-1);
	counts++;
	return(counts-1);	/* returns set ID. */
}


int LSETS :: addmember(int sid,int label){
	if ( label < 0 )  return(-1);	/* label should be >=0 */
	if ( sid >= tblsize )  return(-1);	/* illegal set ID. */
	if ( ! settbl[sid] )  return(-1);	/* the set is not active */
	if ( label >= tblsize_rev ){
		if ( revtbl_realloc(label + 256) )  return(-1);
	}
	if ( revtbl[label] != -1 && revtbl[label] != sid ){
		return(1);	/* the label already belongs to another set */
	}
	if ( 0 > settbl[sid]->addmember(label) )  return(-1);
	revtbl[label] = sid;
	return(0);
}


int LSETS :: mergelabel(int label1,int label2){
	int	i,sid1,sid2;
	if ( label1 >= tblsize_rev || label2 >= tblsize_rev )  return(-1);
	if ( (sid1 = revtbl[label1]) == -1 )  return(-1);
	if ( (sid2 = revtbl[label2]) == -1 )  return(-1);
	if ( settbl[sid1] == 0 || settbl[sid2] == 0 )  return(-1);

	if ( sid1 == sid2 )  return(0);	/* already merged */

	for ( i=0 ; i<settbl[sid2]->counts ; i++ ){
		revtbl[ settbl[sid2]->idlist[i] ] = sid1;
	}
	if ( settbl[sid1]->merge(settbl[sid2]) )  return(-1);
	delete settbl[sid2];
	settbl[sid2] = 0;

	return(0);
}


